@@ -2,7 +2,7 @@
 #ifndef _GLIBMM_CONTAINERHANDLE_SHARED_H
 #define _GLIBMM_CONTAINERHANDLE_SHARED_H
 
-/* $Id: containerhandle_shared.h,v 1.3 2004/02/03 11:08:05 murrayc Exp $ */
+/* $Id: containerhandle_shared.h,v 1.11 2005/01/21 19:26:04 murrayc Exp $ */
 
 /* Copyright (C) 2002 The gtkmm Development Team
  *
@@ -57,8 +57,8 @@ namespace Glib
 enum OwnershipType
 {
   OWNERSHIP_NONE = 0,
-  OWNERSHIP_SHALLOW,
-  OWNERSHIP_DEEP
+  OWNERSHIP_SHALLOW, //Release the list, but not its elements, when the container is deleted
+  OWNERSHIP_DEEP //Release the list, and its elements, when the container is deleted.
 };
 
 
@@ -127,6 +127,10 @@ struct TypeTraits
 
 #ifndef DOXYGEN_SHOULD_SKIP_THIS /* hide the specializations */
 
+//For some (proably, more spec-compliant) compilers, these specializations must
+//be next to the objects that they use.
+#ifdef GLIBMM_CAN_USE_DYNAMIC_CAST_IN_UNUSED_TEMPLATE_WITHOUT_DEFINITION
+
 /** Partial specialization for pointers to GtkObject instances.
  * @ingroup ContHelpers
  */
@@ -139,7 +143,16 @@ struct TypeTraits<T*>
 
   static CType   to_c_type      (CppType ptr) { return Glib::unwrap(ptr);      }
   static CType   to_c_type      (CType   ptr) { return ptr;                    }
-  static CppType to_cpp_type    (CType   ptr) { return Glib::wrap(ptr, false); }
+  static CppType to_cpp_type    (CType   ptr)
+  {
+    //We copy/paste the widget wrap() implementation here,
+    //because we can not use a specific Glib::wrap(T_Impl) overload here,
+    //because that would be "dependent", and g++ 3.4 does not allow that.
+    //The specific Glib::wrap() overloads don't do anything special anyway.
+    GObject* cobj = (GObject*)ptr;
+    return dynamic_cast<CppType>(Glib::wrap_auto(cobj, false /* take_copy */));
+  }
+  
   static void    release_c_type (CType   ptr)
   {
     GLIBMM_DEBUG_UNREFERENCE(0, ptr);
@@ -147,6 +160,9 @@ struct TypeTraits<T*>
   }
 };
 
+//This confuse the SUN Forte compiler, so we ifdef it out:
+#ifdef GLIBMM_HAVE_DISAMBIGUOUS_CONST_TEMPLATE_SPECIALIZATIONS 
+
 /** Partial specialization for pointers to const GtkObject instances.
  * @ingroup ContHelpers
  */
@@ -159,13 +175,23 @@ struct TypeTraits<const T*>
 
   static CType   to_c_type      (CppType ptr) { return Glib::unwrap(ptr);      }
   static CType   to_c_type      (CType   ptr) { return ptr;                    }
-  static CppType to_cpp_type    (CType   ptr) { return Glib::wrap(const_cast<CTypeNonConst>(ptr), false); /* TODO: Allow wrap() to take a const. */}
+  static CppType to_cpp_type    (CType   ptr)
+  {
+     //We copy/paste the widget wrap() implementation here,
+     //because we can not use a specific Glib::wrap(T_Impl) overload here,
+     //because that would be "dependent", and g++ 3.4 does not allow that.
+     //The specific Glib::wrap() overloads don't do anything special anyway.
+     GObject* cobj = (GObject*)const_cast<CTypeNonConst>(ptr);
+     return dynamic_cast<CppType>(Glib::wrap_auto(cobj, false /* take_copy */));
+  }
+  
   static void    release_c_type (CType   ptr)
   {
     GLIBMM_DEBUG_UNREFERENCE(0, ptr);
     g_object_unref(const_cast<CTypeNonConst>(ptr));
   }
 };
+#endif //GLIBMM_HAVE_DISAMBIGUOUS_CONST_TEMPLATE_SPECIALIZATIONS
 
 /** Partial specialization for pointers to GObject instances.
  * @ingroup ContHelpers
@@ -180,7 +206,19 @@ struct TypeTraits< Glib::RefPtr<T> >
 
   static CType   to_c_type      (const CppType& ptr) { return Glib::unwrap(ptr);     }
   static CType   to_c_type      (CType          ptr) { return ptr;                   }
-  static CppType to_cpp_type    (CType          ptr) { return Glib::wrap(ptr, true); }
+  static CppType to_cpp_type    (CType          ptr)
+  {
+    //return Glib::wrap(ptr, true);
+
+    //We copy/paste the wrap() implementation here,
+    //because we can not use a specific Glib::wrap(CType) overload here,
+    //because that would be "dependent", and g++ 3.4 does not allow that.
+    //The specific Glib::wrap() overloads don't do anything special anyway.
+    GObject* cobj = (GObject*)const_cast<CTypeNonConst>(ptr);
+    return Glib::RefPtr<T>( dynamic_cast<T*>(Glib::wrap_auto(cobj, true /* take_copy */)) );
+    //We use dynamic_cast<> in case of multiple inheritance.
+  }
+  
   static void    release_c_type (CType          ptr)
   {
     GLIBMM_DEBUG_UNREFERENCE(0, ptr);
@@ -188,6 +226,9 @@ struct TypeTraits< Glib::RefPtr<T> >
   }
 };
 
+//This confuse the SUN Forte compiler, so we ifdef it out:
+#ifdef GLIBMM_HAVE_DISAMBIGUOUS_CONST_TEMPLATE_SPECIALIZATIONS
+
 /** Partial specialization for pointers to const GObject instances.
  * @ingroup ContHelpers
  * The C++ type is always a Glib::RefPtr<>.
@@ -201,7 +242,19 @@ struct TypeTraits< Glib::RefPtr<const T> >
 
   static CType   to_c_type      (const CppType& ptr) { return Glib::unwrap(ptr);     }
   static CType   to_c_type      (CType          ptr) { return ptr;                   }
-  static CppType to_cpp_type    (CType          ptr) { return Glib::wrap(ptr, true); }
+  static CppType to_cpp_type    (CType          ptr)
+  {
+    //return Glib::wrap(ptr, true);
+
+    //We copy/paste the wrap() implementation here,
+    //because we can not use a specific Glib::wrap(CType) overload here,
+    //because that would be "dependent", and g++ 3.4 does not allow that.
+    //The specific Glib::wrap() overloads don't do anything special anyway.
+    GObject* cobj = (GObject*)(ptr);
+    return Glib::RefPtr<const T>( dynamic_cast<const T*>(Glib::wrap_auto(cobj, true /* take_copy */)) );
+    //We use dynamic_cast<> in case of multiple inheritance.
+  }
+  
   static void    release_c_type (CType          ptr)
   {
     GLIBMM_DEBUG_UNREFERENCE(0, ptr);
@@ -209,6 +262,10 @@ struct TypeTraits< Glib::RefPtr<const T> >
   }
 };
 
+#endif //GLIBMM_HAVE_DISAMBIGUOUS_CONST_TEMPLATE_SPECIALIZATIONS
+
+#endif //GLIBMM_CAN_USE_DYNAMIC_CAST_IN_UNUSED_TEMPLATE_WITHOUT_DEFINITION
+
 /** Specialization for UTF-8 strings.
  * @ingroup ContHelpers
  * When converting from C++ to C, Glib::ustring will be accepted as well as
